#  Copyright (c) 2019. stewu5. All rights reserved.
#  Use of this source code is governed by a BSD-style
#  license that can be found in the LICENSE file.

from .frozen_featurizer_descriptor import FrozenFeaturizerDescriptor
from .mordred_descriptor import Mordred2DDescriptor
